# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import torch
from torchmdnet.models.model import load_model
import warnings

# dict of preset transforms
tranforms = {
    "eV/A -> kcal/mol/A": lambda energy, forces: (
        energy * 23.0609,
        forces * 23.0609,
    ),  # eV->kcal/mol, eV/A -> kcal/mol/A
    "Hartree/Bohr -> kcal/mol/A": lambda energy, forces: (
        energy * 627.509,
        forces * 627.509 / 0.529177,
    ),  # Hartree -> kcal/mol, Hartree/Bohr -> kcal/mol/A
    "Hartree/A -> kcal/mol/A": lambda energy, forces: (
        energy * 627.509,
        forces * 627.509,
    ),  # Hartree -> kcal/mol, Hartree/A -> kcal/mol/A
}


class External:
    """This is an adapter to use TorchMD-Net models in TorchMD.

    Parameters
    ----------
    netfile : str or torch.nn.Module
        Path to the checkpoint file of the model or the model itself.
    embeddings : torch.Tensor
        Embeddings of the atoms in the system.
    device : str, optional
        Device on which the model should be run. Default: "cpu"
    output_transform : str or callable, optional
        Transform to apply to the energy and forces.
        If a string is given, it should be a key in the `transforms` dict.
        If a callable is given, it should take two arguments (energy and forces) and return two tensors of the same shape.
        Default: None
    use_cuda_graph : bool, optional
        Whether to use CUDA graphs to speed up the calculation. Default: False
    cuda_graph_warmup_steps : int, optional
        Number of steps to run as warmup before recording the CUDA graph. Default: 12
    dtype : torch.dtype or str, optional
        Cast the input to this dtype if defined. If passed as a string it should be a valid torch dtype. Default: torch.float32
    kwargs : dict, optional
        Extra arguments to pass to the model when loading it.
    """

    def __init__(
        self,
        netfile,
        embeddings,
        device="cpu",
        output_transform=None,
        use_cuda_graph=False,
        cuda_graph_warmup_steps=12,
        dtype=torch.float32,
        **kwargs,
    ):
        if isinstance(netfile, str):
            extra_args = kwargs
            if use_cuda_graph:
                warnings.warn(
                    "CUDA graphs are enabled, setting static_shapes=True and check_errors=False"
                )
                extra_args["static_shapes"] = True
                extra_args["check_errors"] = False
            self.model = load_model(
                netfile,
                device=device,
                derivative=True,
                **extra_args,
            )
        elif isinstance(netfile, torch.nn.Module):
            if kwargs:
                warnings.warn(
                    "Warning: extra arguments are being ignored when passing a torch.nn.Module"
                )
            self.model = netfile
        else:
            raise ValueError(
                f"Expected a path to a checkpoint file or a torch.nn.Module, got {type(netfile)}"
            )
        self.device = device
        self.n_atoms = embeddings.size(1)
        self.embeddings = embeddings.reshape(-1).to(device)
        self.batch = torch.arange(embeddings.size(0), device=device).repeat_interleave(
            embeddings.size(1)
        )
        self.model.eval()

        if not output_transform:
            self.output_transformer = lambda energy, forces: (
                energy,
                forces,
            )  # identity
        elif output_transform in tranforms.keys():
            self.output_transformer = tranforms[output_transform]
        else:
            self.output_transformer = eval(output_transform)
        if not torch.cuda.is_available() and use_cuda_graph:
            raise ValueError("CUDA graphs are only available if CUDA is")
        self.use_cuda_graph = use_cuda_graph
        self.cuda_graph_warmup_steps = cuda_graph_warmup_steps
        self.cuda_graph = None
        self.energy = None
        self.forces = None
        self.box = None
        self.pos = None
        if isinstance(dtype, str):
            try:
                dtype = getattr(torch, dtype)
            except AttributeError:
                raise ValueError(f"Unknown torch dtype {dtype}")
        self.dtype = dtype

    def _init_cuda_graph(self):
        stream = torch.cuda.Stream()
        self.cuda_graph = torch.cuda.CUDAGraph()
        with torch.cuda.stream(stream):
            for _ in range(self.cuda_graph_warmup_steps):
                self.energy, self.forces = self.model(
                    self.embeddings, self.pos, self.batch, self.box
                )
            with torch.cuda.graph(self.cuda_graph):
                self.energy, self.forces = self.model(
                    self.embeddings, self.pos, self.batch, self.box
                )

    def calculate(self, pos, box=None):
        """Calculate the energy and forces of the system.

        Parameters
        ----------
        pos : torch.Tensor
            Positions of the atoms in the system.
        box : torch.Tensor, optional
            Box vectors of the system. Default: None

        Returns
        -------
        energy : torch.Tensor
            Energy of the system.
        forces : torch.Tensor
            Forces on the atoms in the system.
        """
        pos = pos.to(self.device).to(self.dtype).reshape(-1, 3)
        if box is not None:
            box = box.to(self.device).to(self.dtype)
        if self.use_cuda_graph:
            if self.pos is None:
                self.pos = (
                    pos.clone()
                    .to(self.device)
                    .detach()
                    .requires_grad_(pos.requires_grad)
                )
            if self.box is None and box is not None:
                self.box = box.clone().to(self.device).to(self.dtype).detach()
            if self.cuda_graph is None:
                self._init_cuda_graph()
            assert (
                self.cuda_graph is not None
            ), "CUDA graph is not initialized. This should not had happened."
            with torch.no_grad():
                self.pos.copy_(pos)
                if box is not None:
                    self.box.copy_(box)
                self.cuda_graph.replay()
        else:
            self.energy, self.forces = self.model(self.embeddings, pos, self.batch, box)
        assert self.forces is not None, "The model is not returning forces"
        assert self.energy is not None, "The model is not returning energy"
        return self.output_transformer(
            self.energy.clone().detach(),
            self.forces.clone().reshape(-1, self.n_atoms, 3).detach(),
        )



import ase.calculators.calculator as ase_calc

class TMDNETCalculator(ase_calc.Calculator):
    """This is an adapter to use TorchMD-Net models as an ASE Calculator.

    Parameters
    ----------
    model_file : str
        Path to the checkpoint file of the model.
    device : str, optional
        Device on which the model should be run. Default: "cpu"
    dtype : torch.dtype or str, optional
        Cast the input to this dtype if defined. If passed as a string it should be a valid torch dtype. Default: torch.float32
    compile: bool
        Whether to use torch.compile to speed up calcuations, useful for molecular dynamics. Default: False
    kwargs : dict, optional
        Extra arguments to pass to the model when loading it.
    """

    implemented_properties = [
        "energy",
        "forces"
    ]

    def __init__(
        self,  model_file, device = 'cpu', dtype=torch.float32, compile=False, **kwargs,
    ):
        
        
        ase_calc.Calculator.__init__(self)
        self.device=device
        self.model = load_model(
            model_file,
            derivative=False,
            remove_ref_energy = kwargs.get('remove_ref_energy', True),
            max_num_neighbors = kwargs.get('max_num_neighbors', 64),
            static_shapes = True if compile else False,
            check_errors = False if compile else True,
            **kwargs,
        )
        for parameter in self.model.parameters():
            parameter.requires_grad = False

        self.model.to(device=self.device, dtype=dtype)

        if compile:
            self.compile=True
        else:
            self.compile=False
        
        self.compiled=False

        self.evals = 0



    def calculate(
        self,
        atoms = None,
        properties = None,
        system_changes = ase_calc.all_changes,
    ):


        if not properties:
            properties = ["energy"]


        ase_calc.Calculator.calculate(self, atoms, properties, system_changes)

        numbers = atoms.numbers

        positions = atoms.positions
        total_charge = atoms.info['charge']
        batch = [0 for _ in range(len(numbers))]

        batch = torch.tensor(batch, device=self.device, dtype=torch.long)
        numbers = torch.tensor(numbers, device=self.device, dtype=torch.long)
        positions = torch.tensor(positions, device=self.device, dtype=torch.float32, requires_grad=True)
        total_charge = torch.tensor([total_charge], device=self.device, dtype=torch.float32)

        if self.compile and "numbers" in system_changes:
            if self.evals == 0:
                print("compiling...")
            else:
                print("atomic numbers changed, re-compiling...")

            # Warmup pass to set dim_size before compilation
            # This is needed because torch.compile doesn't support .item() calls
            self.model.to(self.device)
            with torch.no_grad():
                _ = self.model(numbers, positions, batch=batch, q=total_charge)

            self.compiled_model = torch.compile(self.model, backend='inductor', dynamic=False, fullgraph=True, mode='reduce-overhead')
            self.compiled = True


        if self.compiled:
            energy, _ = self.compiled_model(
                numbers, positions, batch=batch, q=total_charge
            )
        else:
            energy, _ = self.model(
                numbers, positions, batch=batch, q=total_charge
            )

        energy.backward()
        forces = -positions.grad

        self.results["energy"] = energy.detach().cpu().item()
        self.results["forces"] = forces.detach().cpu().numpy()
        self.evals +=1 



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import hashlib
from torchmdnet.utils import ATOMIC_NUMBERS
import numpy as np
import os
import torch as pt
from torchmdnet.datasets.memdataset import MemmappedDataset
from torch_geometric.data import Data, download_url
import tarfile
import logging
import re
from tqdm import tqdm


def parse_maceoff_tar(tar_file):
    energy_re = re.compile("energy=(\S+)")
    with tarfile.open(tar_file, "r:gz") as tar:
        for member in tar.getmembers():
            f = tar.extractfile(member)
            if f is None:
                continue
            n_atoms = None
            counter = 0
            positions = []
            numbers = []
            forces = []
            energy = None
            for line in f:
                line = line.decode("utf-8").strip()
                if n_atoms is None:
                    n_atoms = int(line)
                    positions = []
                    numbers = []
                    forces = []
                    energy = None
                    counter = 1
                    continue
                if counter == 1:
                    props = line
                    energy = float(energy_re.search(props).group(1))
                    counter = 2
                    continue
                el, x, y, z, fx, fy, fz, _, _, _ = line.split()
                numbers.append(ATOMIC_NUMBERS[el])
                positions.append([float(x), float(y), float(z)])
                forces.append([float(fx), float(fy), float(fz)])
                counter += 1
                if counter == n_atoms + 2:
                    n_atoms = None
                    yield energy, numbers, positions, forces


class MACEOFF(MemmappedDataset):
    """
        MACEOFF dataset from MACE-OFF23: Transferable Machine Learning Force Fields for Organic Molecules, Kovacs et.al. https://arxiv.org/abs/2312.15211
        This dataset consists of arounf 100K conformations with 95% of them coming from SPICE and augmented with conformations from QMugs, COMP6 and clusters of water carved out of MD simulations of liquid water.

        From the repository:
        The core of the training set is the SPICE dataset. 95% of the data were used for training and validation, and 5% for testing. The MACE-OFF23 model is trained to reproduce the energies and forces computed at the ωB97M-D3(BJ)/def2-TZVPPD level of quantum mechanics, as implemented in the PSI4 software. We have used a subset of SPICE that contains the ten chemical elements H, C, N, O, F, P, S, Cl, Br, and I, and has a neutral formal charge. We have also removed the ion pairs subset. Overall, we used about 85% of the full SPICE dataset.

    Contains energy and force data in units of eV and eV/Angstrom

    """

    VERSIONS = {
        "1.0": {
            "url": "https://api.repository.cam.ac.uk/server/api/core/bitstreams/b185b5ab-91cf-489a-9302-63bfac42824a/content",
            "file": "train_large_neut_no_bad_clean.tar.gz",
        },
    }

    @property
    def raw_dir(self):
        return os.path.join(super().raw_dir, "maceoff", self.version)

    @property
    def raw_file_names(self):
        return self.VERSIONS[self.version]["file"]

    @property
    def raw_url(self):
        return f"{self.VERSIONS[self.version]['url']}"

    def __init__(
        self,
        root=None,
        transform=None,
        pre_transform=None,
        pre_filter=None,
        version="1.0",
        max_gradient=None,
    ):
        arg_hash = f"{version}{max_gradient}"
        arg_hash = hashlib.md5(arg_hash.encode()).hexdigest()
        self.name = f"{self.__class__.__name__}-{arg_hash}"
        self.version = str(version)
        assert self.version in self.VERSIONS
        self.max_gradient = max_gradient
        super().__init__(
            root,
            transform,
            pre_transform,
            pre_filter,
            properties=("y", "neg_dy"),
        )

    def sample_iter(self, mol_ids=False):
        assert len(self.raw_paths) == 1
        logging.info(f"Processing dataset {self.raw_file_names}")
        for energy, numbers, positions, forces in tqdm(
            parse_maceoff_tar(self.raw_paths[0]), desc="Processing conformations"
        ):
            data = Data(
                **dict(
                    z=pt.tensor(np.array(numbers), dtype=pt.long),
                    pos=pt.tensor(positions, dtype=pt.float32),
                    y=pt.tensor(energy, dtype=pt.float64).view(1, 1),
                    neg_dy=pt.tensor(forces, dtype=pt.float32),
                )
            )
            assert data.y.shape == (1, 1)
            assert data.z.shape[0] == data.pos.shape[0]
            assert data.neg_dy.shape[0] == data.pos.shape[0]
            # Skip samples with large forces
            if self.max_gradient:
                if data.neg_dy.norm(dim=1).max() > float(self.max_gradient):
                    continue
            if self.pre_filter is not None and not self.pre_filter(data):
                continue
            if self.pre_transform is not None:
                data = self.pre_transform(data)
            yield data

    def download(self):
        download_url(self.raw_url, self.raw_dir, filename=self.raw_file_names)



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

from torchmdnet.priors.atomref import Atomref, LearnableAtomref
from torchmdnet.priors.d2 import D2
from torchmdnet.priors.zbl import ZBL
from torchmdnet.priors.coulomb import Coulomb

__all__ = ["Atomref", "LearnableAtomref", "D2", "ZBL", "Coulomb"]



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import torch
from torch_geometric.data import Dataset, Data
import h5py
import numpy as np


class HDF5(Dataset):
    """A custom dataset that loads data from a HDF5 file.

    To use this, dataset_root should be  the path to the HDF5 file, or
    alternatively a semicolon separated  list of multiple files.  Each
    group in the  file contains samples that all have  the same number
    of atoms.  Typically there is one  group for each unique number of
    atoms, but that is not required.  Each group should contain arrays
    called "types" (atom type indices), "pos" (atom positions), and at
    least one of "energy" (the  energy of each sample) and/or "forces"
    (the force on each atom).

    Args:
        filename (string): A semicolon separated list of HDF5 files.
        dataset_preload_limit (int, optional): If the dataset is smaller than this limit (in MB), preload it into CPU memory. (default: :obj:`1024`)

    """

    def __init__(self, filename, dataset_preload_limit=1024, **kwargs):
        super(HDF5, self).__init__()
        self.filename = filename
        self.index = None
        self.fields = None
        self.num_molecules = 0
        files = [h5py.File(f, "r") for f in self.filename.split(";")]
        total_file_size = sum([f.id.get_filesize() for f in files])
        print(f"Loading {len(files)} HDF5 files ({total_file_size / 1024**2:.2f} MB)")
        for file in files:
            for group_name in file:
                group = file[group_name]
                if group_name == "_metadata":
                    for name in group:
                        setattr(self, name, torch.tensor(np.array(group[name])))
                else:
                    self.num_molecules += len(group["pos"])
                    if self.fields is None:
                        # Record which data fields are present in this file.
                        self.fields = [
                            ("pos", "pos", torch.float32),
                            ("z", "types", torch.long),
                        ]
                        if "energy" in group:
                            self.fields.append(("y", "energy", torch.float32))
                        if "forces" in group:
                            self.fields.append(("neg_dy", "forces", torch.float32))
                        if "partial_charges" in group:
                            self.fields.append(
                                ("partial_charges", "partial_charges", torch.float32)
                            )
                        assert ("energy" in group) or (
                            "forces" in group
                        ), "Each group must contain at least energies or forces"
            file.close()
        self.cached = False
        if total_file_size <= dataset_preload_limit * 1024**2:
            print(
                f"Preloading {len(files)} HDF5 files ({total_file_size / 1024**2:.2f} MB)"
            )
            self.cached = True
            self._preload_data()

    def _preload_data(self):
        """Preload the entire dataset into memory.
        Store it in a dictionary of torch tensors. The dictionary has an entry for each field and group.
        """
        self.stored_data = {}
        for field in self.fields:
            self.stored_data[field] = []
        self.index = []
        files = [h5py.File(f, "r") for f in self.filename.split(";")]
        i = 0
        for file in files:
            for group_name, group in file.items():
                if group_name != "_metadata":
                    size = len(group["pos"])
                    for field in self.fields:
                        data = group[field[1]]
                        dtype = field[2]
                        # Watchout for the 1D case, embed can be shared for all samples
                        tmp = torch.tensor(np.array(data), dtype=dtype)
                        if tmp.ndim == 1:
                            if len(tmp) == size:
                                tmp = tmp.unsqueeze(-1)
                            else:
                                tmp = tmp.unsqueeze(0).expand(size, -1)
                        self.stored_data[field].append(tmp)
                    self.index.extend(list(zip([i] * size, range(size))))
                    i += 1
            file.close()

    def _setup_index(self):
        files = [h5py.File(f, "r") for f in self.filename.split(";")]
        self.index = []
        for file in files:
            for group_name, group in file.items():
                if group_name != "_metadata":
                    data = [group[field[1]] for field in self.fields]
                    self.index.extend(
                        [tuple(data + [i]) for i in range(len(group["pos"]))]
                    )
        assert self.num_molecules == len(
            self.index
        ), f"Mismatch between previously calculated molecule count ({self.num_molecules}) and actual molecule count ({len(self.index)})"

    def get(self, idx):
        data = Data()
        if self.cached:
            # If the dataset is cached, the index is just a list of indices into the stored data.
            fileid, index = self.index[idx]
            for field in self.fields:
                data[field[0]] = self.stored_data[field][fileid][index]
        else:
            # only open files here to avoid copying objects of this class to another
            # process with open file handles (potentially corrupts h5py loading)
            if self.index is None:
                self._setup_index()
            *fields_data, i = self.index[idx]
            # Assuming the first element of fields_data is 'pos' based on the definition of self.fields
            size = len(fields_data[0])
            for (name, _, dtype), d in zip(self.fields, fields_data):
                if d.ndim == 1:
                    tensor_input = [d[i]] if len(d) == size else d[:]
                else:
                    tensor_input = d[i]
                data[name] = torch.tensor(tensor_input, dtype=dtype)
        return data

    def len(self):
        return self.num_molecules



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

from abc import abstractmethod, ABCMeta
from typing import Optional, Tuple
from torch import nn, Tensor


class BaseWrapper(nn.Module, metaclass=ABCMeta):
    r"""Base class for model wrappers.

    Children of this class should implement the `forward` method,
    which calls `self.model(z, pos, batch=batch)` at some point.
    Wrappers that are applied before the REDUCE operation should return
    the model's output, `z`, `pos`, `batch` and potentially vector
    features`v`. Wrappers that are applied after REDUCE should only
    return the model's output.
    """

    def __init__(self, model):
        super(BaseWrapper, self).__init__()
        self.model = model

    def reset_parameters(self):
        self.model.reset_parameters()

    @abstractmethod
    def forward(self, z, pos, batch=None):
        return


class AtomFilter(BaseWrapper):
    """
    Remove atoms with Z > remove_threshold from the model's output.
    """

    def __init__(self, model, remove_threshold):
        super(AtomFilter, self).__init__(model)
        self.remove_threshold = remove_threshold

    def forward(
        self,
        z: Tensor,
        pos: Tensor,
        batch: Tensor,
        q: Optional[Tensor] = None,
        s: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Optional[Tensor], Tensor, Tensor, Tensor]:
        x, v, z, pos, batch = self.model(z, pos, batch=batch, q=q, s=s)

        n_samples = len(batch.unique())

        # drop atoms according to the filter
        atom_mask = z > self.remove_threshold
        x = x[atom_mask]
        if v is not None:
            v = v[atom_mask]
        z = z[atom_mask]
        pos = pos[atom_mask]
        batch = batch[atom_mask]

        assert len(batch.unique()) == n_samples, (
            "Some samples were completely filtered out by the atom filter. "
            f"Make sure that at least one atom per sample exists with Z > {self.remove_threshold}."
        )
        return x, v, z, pos, batch



from torch_geometric.data import InMemoryDataset, Data
import torch
import numpy as np
import os
import requests
import zipfile
import re


def create_numpy_arrays(file_path):
    with open(file_path, "r") as file:
        num_atoms = int(file.readline().strip())
        file.seek(0)
        num_conformations = sum(1 for line in file if line.strip().isdigit())
        file.seek(0)
        energies = np.zeros((num_conformations, 1))
        forces = np.zeros((num_conformations, num_atoms, 3))
        positions = np.zeros((num_conformations, num_atoms, 3))
        atomic_numbers = np.zeros((num_conformations, num_atoms, 1), dtype=int)
        box_vectors = np.zeros((num_conformations, 9))
        # Extracting TotEnergy, pbc, and Lattice
        for i in range(num_conformations):
            _ = file.readline()
            properties_line = file.readline()
            tot_energy_match = re.search(r"TotEnergy=(-?\d+\.\d+)", properties_line)
            pbc_match = re.search(r'pbc="([T|F] [T|F] [T|F])"', properties_line)
            lattice_match = re.search(r'Lattice="([-?\d+.\d+\s]+)"', properties_line)
            energies[i] = float(tot_energy_match.group(1)) if tot_energy_match else None
            pbc = [s == "T" for s in pbc_match.group(1).split()] if pbc_match else None
            assert pbc == [True, True, True] or pbc == [False, False, False]
            box_vectors[i] = (
                [float(x) for x in lattice_match.group(1).split()]
                if lattice_match
                else None
            )
            for j in range(num_atoms):
                atom_line = file.readline().strip().split()
                positions[i, j] = [float(x) for x in atom_line[1:4]]
                forces[i, j] = [float(x) for x in atom_line[4:7]]
                atomic_numbers[i, j] = int(atom_line[7])
    return energies, forces, positions, atomic_numbers, box_vectors


class WaterBox(InMemoryDataset):
    """WaterBox dataset from [1]_.

    The dataset consists of 1593 water molecules in a cubic box with periodic boundary conditions.
    The molecules are sampled from a molecular dynamics simulation of liquid water.

    Each sample in the dataset contains the following properties:

    - z (LongTensor): Atomic numbers of the atoms in the molecule.
    - pos (FloatTensor): Positions of the atoms in the molecule.
    - y (FloatTensor): Total energy of the molecule.
    - neg_dy (FloatTensor): Negative of the forces on the atoms in the molecule.
    - box (FloatTensor): Box vectors of the simulation cell.

    Parameters
    ----------
    root : str
        Root directory where the dataset should be saved.

    References
    ----------
    [1] Ab initio thermodynamics of liquid and solid water. Bingqing et. al. https://arxiv.org/abs/1811.08630
    """

    url = "https://archive.materialscloud.org/record/file?record_id=71&filename=training-set.zip"

    def __init__(self, root, transform=None, pre_transform=None):
        super(WaterBox, self).__init__(root, transform, pre_transform)
        self.load(self.processed_paths[0])

    @property
    def raw_file_names(self):
        return ["dataset_1593.xyz"]

    @property
    def processed_file_names(self):
        return ["data.pt"]

    def download(self):
        r = requests.get(self.url)
        if r.status_code != 200:
            raise Exception(
                f"Failed to download file from {self.url}. Status code: {r.status_code}"
            )
        zip_path = os.path.join(self.raw_dir, "training-set.zip")
        with open(zip_path, "wb") as f:
            f.write(r.content)
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(self.raw_dir)

    def process(self):
        dataset_xyz_path = os.path.join(
            self.raw_dir, "training-set", "dataset_1593.xyz"
        )
        energies, forces, positions, atomic_numbers, box_vectors = create_numpy_arrays(
            dataset_xyz_path
        )

        data_list = []
        for i in range(len(energies)):
            z = torch.tensor(atomic_numbers[i], dtype=torch.long).view(-1)
            pos = torch.tensor(positions[i], dtype=torch.float)
            y = torch.tensor(energies[i], dtype=torch.float)
            neg_dy = torch.tensor(forces[i], dtype=torch.float)
            box = torch.tensor(box_vectors[i], dtype=torch.float).view(1, 3, 3)
            data_list.append(Data(z=z, pos=pos, y=y, neg_dy=neg_dy, box=box))

        self.save(data_list, self.processed_paths[0])



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

from collections import defaultdict
import torch
from torch.optim import AdamW
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.nn.functional import local_response_norm
from torch import Tensor
from typing import Optional, Dict, Tuple
from lightning import LightningModule
from torchmdnet.models.model import create_model, load_model
from torchmdnet.models.utils import dtype_mapping
from torchmdnet.loss import l1_loss, loss_class_mapping
import torch_geometric.transforms as T


class FloatCastDatasetWrapper(T.BaseTransform):
    """A transform that casts all floating point tensors to a given dtype.
    tensors to a given dtype.
    """

    def __init__(self, dtype=torch.float64):
        super(FloatCastDatasetWrapper, self).__init__()
        self._dtype = dtype

    def forward(self, data):
        for key, value in data:
            if torch.is_tensor(value) and torch.is_floating_point(value):
                setattr(data, key, value.to(self._dtype))
        return data


class EnergyRefRemover(T.BaseTransform):
    """A transform that removes the atom reference energy from the energy of a
    dataset.
    """

    def __init__(self, atomref):
        super(EnergyRefRemover, self).__init__()
        self._atomref = atomref

    def forward(self, data):
        self._atomref = self._atomref.to(data.z.device).type(data.y.dtype)
        if "y" in data:
            data.y.index_add_(0, data.batch, -self._atomref[data.z])
        return data


# This wrapper is here in order to permit Lightning to serialize the loss function.
class LossFunction:
    def __init__(self, loss_fn, extra_args=None):
        self.loss_fn = loss_fn
        self.extra_args = extra_args
        if self.extra_args is None:
            self.extra_args = {}

    def __call__(self, x, batch):
        return self.loss_fn(x, batch, **self.extra_args)


class LNNP(LightningModule):
    """
    Lightning wrapper for the Neural Network Potentials in TorchMD-Net.

    Args:
        hparams (dict): A dictionary containing the hyperparameters of the model.
        prior_model (torchmdnet.priors.BasePrior): A prior model to use in the model.
        mean (torch.Tensor, optional): The mean of the dataset to normalize the input.
        std (torch.Tensor, optional): The standard deviation of the dataset to normalize the input.
    """

    def __init__(self, hparams, prior_model=None, mean=None, std=None):
        super(LNNP, self).__init__()
        if "charge" not in hparams:
            hparams["charge"] = False
        if "spin" not in hparams:
            hparams["spin"] = False
        if "train_loss" not in hparams:
            hparams["train_loss"] = "mse_loss"
        if "train_loss_arg" not in hparams:
            hparams["train_loss_arg"] = {}
        self.save_hyperparameters(hparams)

        if self.hparams.load_model:
            self.model = load_model(self.hparams.load_model, args=self.hparams)
        else:
            self.model = create_model(self.hparams, prior_model, mean, std)

        # initialize exponential smoothing
        self.ema = None
        self._reset_ema_dict()

        # initialize loss collection
        self.losses = None
        self._reset_losses_dict()

        self.data_transform = FloatCastDatasetWrapper(
            dtype_mapping[self.hparams.precision]
        )
        if self.hparams.remove_ref_energy:
            self.data_transform = T.Compose(
                [
                    EnergyRefRemover(self.model.prior_model[-1].initial_atomref),
                    self.data_transform,
                ]
            )

        if self.hparams.train_loss not in loss_class_mapping:
            raise ValueError(
                f"Training loss {self.hparams.train_loss} not supported. Supported losses are {list(loss_class_mapping.keys())}"
            )

        self.train_loss_fn = LossFunction(
            loss_class_mapping[self.hparams.train_loss],
            self.hparams.train_loss_arg,
        )

    def configure_optimizers(self):
        optimizer = AdamW(
            self.model.parameters(),
            lr=self.hparams.lr,
            weight_decay=self.hparams.weight_decay,
        )
        scheduler = ReduceLROnPlateau(
            optimizer,
            "min",
            factor=self.hparams.lr_factor,
            patience=self.hparams.lr_patience,
            min_lr=self.hparams.lr_min,
        )
        lr_metric = getattr(self.hparams, "lr_metric", "val")
        monitor = f"{lr_metric}_total_{self.hparams.train_loss}"
        lr_scheduler = {
            "scheduler": scheduler,
            "strict": True,
            "monitor": monitor,
            "interval": "epoch",
            "frequency": 1,
        }
        return [optimizer], [lr_scheduler]

    def forward(
        self,
        z: Tensor,
        pos: Tensor,
        batch: Optional[Tensor] = None,
        box: Optional[Tensor] = None,
        q: Optional[Tensor] = None,
        s: Optional[Tensor] = None,
        extra_args: Optional[Dict[str, Tensor]] = None,
    ) -> Tuple[Tensor, Optional[Tensor]]:
        return self.model(z, pos, batch=batch, box=box, q=q, s=s, extra_args=extra_args)

    def training_step(self, batch, batch_idx):
        return self.step(
            batch, [(self.hparams.train_loss, self.train_loss_fn)], "train"
        )

    def validation_step(self, batch, batch_idx, *args):
        # If args is not empty the first (and only) element is the dataloader_idx
        # We want to test every number of epochs just for reporting, but this is not supported by Lightning.
        # Instead, we trick it by providing two validation dataloaders and interpreting the second one as test.
        # The dataloader takes care of sending the two sets only when the second one is needed.
        is_val = len(args) == 0 or (len(args) > 0 and args[0] == 0)
        if is_val:
            step_type = {
                "loss_fn_list": [
                    ("l1_loss", l1_loss),
                    (self.hparams.train_loss, self.train_loss_fn),
                ],
                "stage": "val",
            }
        else:
            step_type = {"loss_fn_list": [("l1_loss", l1_loss)], "stage": "test"}
        return self.step(batch, **step_type)

    def test_step(self, batch, batch_idx):
        return self.step(batch, [("l1_loss", l1_loss)], "test")

    def predict_step(self, batch, batch_idx):
        batch = self.data_transform(batch)

        with torch.set_grad_enabled(self.hparams.derivative):
            extra_args = batch.to_dict()
            for a in ("y", "neg_dy", "z", "pos", "batch", "box", "q", "s"):
                if a in extra_args:
                    del extra_args[a]
            return self(
                batch.z,
                batch.pos,
                batch=batch.batch,
                box=batch.box if "box" in batch else None,
                q=batch.q if self.hparams.charge else None,
                s=batch.s if self.hparams.spin else None,
                extra_args=extra_args,
            )

    def _compute_losses(self, y, neg_y, batch, loss_fn, loss_name, stage):
        # Compute the loss for the predicted value and the negative derivative (if available)
        # Args:
        #   y: predicted value
        #   neg_y: predicted negative derivative
        #   batch: batch of data
        #   loss_fn: The loss function to compute
        #   loss_name: The name of the loss function
        # Returns:
        #   loss_y: loss for the predicted value
        #   loss_neg_y: loss for the predicted negative derivative
        loss_y, loss_neg_y = torch.tensor(0.0, device=self.device), torch.tensor(
            0.0, device=self.device
        )
        if self.hparams.derivative and "neg_dy" in batch:
            loss_neg_y = loss_fn(neg_y, batch.neg_dy)
            loss_neg_y = self._update_loss_with_ema(
                stage, "neg_dy", loss_name, loss_neg_y
            )
        if "y" in batch:
            loss_y = loss_fn(y, batch.y)
            loss_y = self._update_loss_with_ema(stage, "y", loss_name, loss_y)
        return {"y": loss_y, "neg_dy": loss_neg_y}

    def _update_loss_with_ema(self, stage, type, loss_name, loss):
        # Update the loss using an exponential moving average when applicable
        # Args:
        #   stage: stage of the training (train, val, test)
        #   type: type of loss (y, neg_dy)
        #   loss_name: name of the loss function
        #   loss: loss value
        alpha = getattr(self.hparams, f"ema_alpha_{type}")
        if stage in ["train", "val"] and alpha < 1 and alpha > 0:
            ema = (
                self.ema[stage][type][loss_name]
                if loss_name in self.ema[stage][type]
                else loss.detach()
            )
            loss = alpha * loss + (1 - alpha) * ema
            self.ema[stage][type][loss_name] = loss.detach()
        return loss

    def step(self, batch, loss_fn_list, stage):
        # Run a forward pass and compute the loss for each loss function
        # If the batch contains the derivative, also compute the loss for the negative derivative
        # Args:
        #   batch: batch of data
        #   loss_fn_list: list of loss functions to compute and record (the last one is used for the total loss returned by this function)
        #   stage: stage of the training (train, val, test)
        # Returns:
        #   total_loss: sum of all losses (weighted by the loss weights) for the last loss function in the provided list
        assert len(loss_fn_list) > 0
        assert self.losses is not None
        batch = self.data_transform(batch)
        with torch.set_grad_enabled(stage == "train" or self.hparams.derivative):
            extra_args = batch.to_dict()
            for a in ("y", "neg_dy", "z", "pos", "batch", "box", "q", "s"):
                if a in extra_args:
                    del extra_args[a]
            # TODO: the model doesn't necessarily need to return a derivative once
            # Union typing works under TorchScript (https://github.com/pytorch/pytorch/pull/53180)
            y, neg_dy = self(
                batch.z,
                batch.pos,
                batch=batch.batch,
                box=batch.box if "box" in batch else None,
                q=batch.q if self.hparams.charge else None,
                s=batch.s if self.hparams.spin else None,
                extra_args=extra_args,
            )
        if self.hparams.derivative and "y" not in batch:
            # "use" both outputs of the model's forward function but discard the first
            # to only use the negative derivative and avoid 'Expected to have finished reduction
            # in the prior iteration before starting a new one.', which otherwise get's
            # thrown because of setting 'find_unused_parameters=False' in the DDPPlugin
            neg_dy = neg_dy + y.sum() * 0
        if "y" in batch and batch.y.ndim == 1:
            batch.y = batch.y.unsqueeze(1)
        for loss_name, loss_fn in loss_fn_list:
            step_losses = self._compute_losses(
                y, neg_dy, batch, loss_fn, loss_name, stage
            )
            if self.hparams.neg_dy_weight > 0:
                self.losses[stage]["neg_dy"][loss_name].append(
                    step_losses["neg_dy"].detach()
                )
            if self.hparams.y_weight > 0:
                self.losses[stage]["y"][loss_name].append(step_losses["y"].detach())
            total_loss = (
                step_losses["y"] * self.hparams.y_weight
                + step_losses["neg_dy"] * self.hparams.neg_dy_weight
            )
            self.losses[stage]["total"][loss_name].append(total_loss.detach())
        return total_loss

    def optimizer_step(self, *args, **kwargs):
        optimizer = kwargs["optimizer"] if "optimizer" in kwargs else args[2]
        if self.trainer.global_step < self.hparams.lr_warmup_steps:
            lr_scale = min(
                1.0,
                float(self.trainer.global_step + 1)
                / float(self.hparams.lr_warmup_steps),
            )

            for pg in optimizer.param_groups:
                pg["lr"] = lr_scale * self.hparams.lr
        super().optimizer_step(*args, **kwargs)
        optimizer.zero_grad()

    def _get_mean_loss_dict_for_type(self, type):
        # Returns a list with the mean loss for each loss_fn for each stage (train, val, test)
        # Parameters:
        # type: either y, neg_dy or total
        # Returns:
        # A dict with an entry for each stage (train, val, test) with the mean loss for each loss_fn (e.g. mse_loss)
        # The key for each entry is "stage_type_loss_fn"
        assert self.losses is not None
        mean_losses = {}
        for stage in ["train", "val", "test"]:
            for loss_fn_name in self.losses[stage][type].keys():
                mean_losses[stage + "_" + type + "_" + loss_fn_name] = torch.stack(
                    self.losses[stage][type][loss_fn_name]
                ).mean()
        return mean_losses

    def on_validation_epoch_end(self):
        if not self.trainer.sanity_checking:
            # construct dict of logged metrics
            result_dict = {
                "epoch": float(self.current_epoch),
                "lr": self.trainer.optimizers[0].param_groups[0]["lr"],
            }
            result_dict.update(self._get_mean_loss_dict_for_type("total"))
            result_dict.update(self._get_mean_loss_dict_for_type("y"))
            result_dict.update(self._get_mean_loss_dict_for_type("neg_dy"))
            self.log_dict(result_dict, sync_dist=True)

        self._reset_losses_dict()

    def on_test_epoch_end(self):
        # Log all test losses
        if not self.trainer.sanity_checking:
            result_dict = {}
            result_dict.update(self._get_mean_loss_dict_for_type("total"))
            result_dict.update(self._get_mean_loss_dict_for_type("y"))
            result_dict.update(self._get_mean_loss_dict_for_type("neg_dy"))
            # Get only test entries
            result_dict = {k: v for k, v in result_dict.items() if k.startswith("test")}
            self.log_dict(result_dict, sync_dist=True)

    def on_train_epoch_end(self):
        # Log all train losses
        if not self.trainer.sanity_checking:
            result_dict = {}
            result_dict.update(self._get_mean_loss_dict_for_type("total"))
            result_dict.update(self._get_mean_loss_dict_for_type("y"))
            result_dict.update(self._get_mean_loss_dict_for_type("neg_dy"))
            # Get only train entries
            result_dict = {
                k: v for k, v in result_dict.items() if k.startswith("train")
            }
            self.log_dict(result_dict, sync_dist=True)

    def _reset_losses_dict(self):
        # Losses has an entry for each stage in ["train", "val", "test"]
        # Each entry has an entry with "total", "y" and "neg_dy"
        # Each of these entries has an entry for each loss_fn (e.g. mse_loss)
        # The loss_fn values are not known in advance
        self.losses = {}
        for stage in ["train", "val", "test"]:
            self.losses[stage] = {}
            for loss_type in ["total", "y", "neg_dy"]:
                self.losses[stage][loss_type] = defaultdict(list)

    def _reset_ema_dict(self):
        self.ema = {}
        for stage in ["train", "val"]:
            self.ema[stage] = {}
            for loss_type in ["y", "neg_dy"]:
                self.ema[stage][loss_type] = {}



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

from abc import abstractmethod, ABCMeta
from typing import Optional
import torch
from torch import nn
from torchmdnet.models.utils import GatedEquivariantBlock, scatter, MLP
from torchmdnet.utils import atomic_masses
from warnings import warn

__all__ = ["Scalar", "DipoleMoment", "ElectronicSpatialExtent"]


class OutputModel(nn.Module, metaclass=ABCMeta):
    """Base class for output models.

    Derive this class to make custom output models.
    As an example, have a look at the :py:mod:`torchmdnet.output_modules.Scalar` output model.
    """

    def __init__(self, allow_prior_model, reduce_op, static_shapes=False):
        super(OutputModel, self).__init__()
        self.allow_prior_model = allow_prior_model
        self.reduce_op = reduce_op
        self.static_shapes = static_shapes
        self.dim_size = 0

    def reset_parameters(self):
        pass

    @abstractmethod
    def pre_reduce(self, x, v, z, pos, batch):
        return

    def reduce(self, x, batch):
        # torch.compile and torch.export don't support .item() calls during tracing
        # The model should be warmed up before compilation to set the correct dim_size
        if torch.compiler.is_compiling():
            pass
        elif torch.jit.is_scripting():
            # TorchScript doesn't support torch.cuda.is_current_stream_capturing()
            # For CPU, always update dim_size (no CUDA graphs on CPU)
            # For CUDA with static_shapes, only update once (first call sets dim_size for CUDA graph capture)
            # For CUDA without static_shapes, always update (dynamic batch sizes)
            if not x.is_cuda or not self.static_shapes or self.dim_size == 0:
                self.dim_size = int(batch.max().item() + 1)
        else:
            is_capturing = x.is_cuda and torch.cuda.is_current_stream_capturing()
            if not x.is_cuda or not is_capturing:
                self.dim_size = int(batch.max().item() + 1)
            if is_capturing:
                assert (
                    self.dim_size > 0
                ), "Warming up is needed before capturing the model into a CUDA graph"
                warn(
                    "CUDA graph capture will lock the batch to the current number of samples ({}). Changing this will result in a crash".format(
                        self.dim_size
                    )
                )
        return scatter(x, batch, dim=0, dim_size=self.dim_size, reduce=self.reduce_op)

    def post_reduce(self, x):
        return x


class Scalar(OutputModel):
    def __init__(
        self,
        hidden_channels,
        activation="silu",
        allow_prior_model=True,
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(Scalar, self).__init__(
            allow_prior_model=allow_prior_model,
            reduce_op=reduce_op,
            static_shapes=static_shapes,
        )
        self.output_network = MLP(
            in_channels=hidden_channels,
            out_channels=1,
            hidden_channels=hidden_channels // 2,
            activation=activation,
            num_hidden_layers=kwargs.get("num_layers", 0),
            dtype=dtype,
        )
        self.reset_parameters()

    def reset_parameters(self):
        self.output_network.reset_parameters()

    def pre_reduce(self, x, v: Optional[torch.Tensor], z, pos, batch):
        return self.output_network(x)


class EquivariantScalar(OutputModel):
    def __init__(
        self,
        hidden_channels,
        activation="silu",
        allow_prior_model=True,
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(EquivariantScalar, self).__init__(
            allow_prior_model=allow_prior_model,
            reduce_op=reduce_op,
            static_shapes=static_shapes,
        )
        if kwargs.get("num_layers", 0) > 0:
            warn("num_layers is not used in EquivariantScalar")
        self.output_network = nn.ModuleList(
            [
                GatedEquivariantBlock(
                    hidden_channels,
                    hidden_channels // 2,
                    activation=activation,
                    scalar_activation=True,
                    dtype=dtype,
                ),
                GatedEquivariantBlock(
                    hidden_channels // 2, 1, activation=activation, dtype=dtype
                ),
            ]
        )

        self.reset_parameters()

    def reset_parameters(self):
        for layer in self.output_network:
            layer.reset_parameters()

    def pre_reduce(self, x, v, z, pos, batch):
        for layer in self.output_network:
            x, v = layer(x, v)
        # include v in output to make sure all parameters have a gradient
        return x + v.sum() * 0


class DipoleMoment(Scalar):
    def __init__(
        self,
        hidden_channels,
        activation="silu",
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(DipoleMoment, self).__init__(
            hidden_channels,
            activation,
            allow_prior_model=False,
            reduce_op=reduce_op,
            dtype=dtype,
            static_shapes=static_shapes,
            **kwargs,
        )
        atomic_mass = torch.from_numpy(atomic_masses).to(dtype)
        self.register_buffer("atomic_mass", atomic_mass)

    def pre_reduce(self, x, v: Optional[torch.Tensor], z, pos, batch):
        x = self.output_network(x)

        # Get center of mass.
        mass = self.atomic_mass[z].view(-1, 1)
        c = scatter(mass * pos, batch, dim=0) / scatter(mass, batch, dim=0)
        x = x * (pos - c[batch])
        return x

    def post_reduce(self, x):
        return torch.norm(x, dim=-1, keepdim=True)


class EquivariantDipoleMoment(EquivariantScalar):
    def __init__(
        self,
        hidden_channels,
        activation="silu",
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(EquivariantDipoleMoment, self).__init__(
            hidden_channels,
            activation,
            allow_prior_model=False,
            reduce_op=reduce_op,
            dtype=dtype,
            static_shapes=static_shapes,
            **kwargs,
        )
        atomic_mass = torch.from_numpy(atomic_masses).to(dtype)
        self.register_buffer("atomic_mass", atomic_mass)

    def pre_reduce(self, x, v, z, pos, batch):
        for layer in self.output_network:
            x, v = layer(x, v)

        # Get center of mass.
        mass = self.atomic_mass[z].view(-1, 1)
        c = scatter(mass * pos, batch, dim=0) / scatter(mass, batch, dim=0)
        x = x * (pos - c[batch])
        return x + v.squeeze()

    def post_reduce(self, x):
        return torch.norm(x, dim=-1, keepdim=True)


class ElectronicSpatialExtent(OutputModel):
    def __init__(
        self,
        hidden_channels,
        activation="silu",
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(ElectronicSpatialExtent, self).__init__(
            allow_prior_model=False, reduce_op=reduce_op, static_shapes=static_shapes
        )
        self.output_network = MLP(
            in_channels=hidden_channels,
            out_channels=1,
            hidden_channels=hidden_channels // 2,
            activation=activation,
            num_hidden_layers=kwargs.get("num_layers", 0),
            dtype=dtype,
        )
        atomic_mass = torch.from_numpy(atomic_masses).to(dtype)
        self.register_buffer("atomic_mass", atomic_mass)

        self.reset_parameters()

    def reset_parameters(self):
        self.output_network.reset_parameters()

    def pre_reduce(self, x, v: Optional[torch.Tensor], z, pos, batch):
        x = self.output_network(x)

        # Get center of mass.
        mass = self.atomic_mass[z].view(-1, 1)
        c = scatter(mass * pos, batch, dim=0) / scatter(mass, batch, dim=0)

        x = torch.norm(pos - c[batch], dim=1, keepdim=True) ** 2 * x
        return x


class EquivariantElectronicSpatialExtent(ElectronicSpatialExtent):
    pass


class EquivariantVectorOutput(EquivariantScalar):
    def __init__(
        self,
        hidden_channels,
        activation="silu",
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(EquivariantVectorOutput, self).__init__(
            hidden_channels,
            activation,
            allow_prior_model=False,
            reduce_op="sum",
            dtype=dtype,
            static_shapes=static_shapes,
            **kwargs,
        )

    def pre_reduce(self, x, v, z, pos, batch):
        for layer in self.output_network:
            x, v = layer(x, v)
        return v.squeeze()


def _exp_cutoff(d, rc: float):
    """short range cutoff to damp close range coulomb interaction

    Implementation copied form aimnetcentral
    """
    fc = (
        torch.exp(-1.0 / (1.0 - (d / rc).clamp(0, 1.0 - 1e-6).pow(2)))
        / 0.36787944117144233
    )
    return fc


def _triu_indices(x, batch):
    N = x.shape[0]
    i, j = torch.triu_indices(N, N, 1, device=x.device, dtype=torch.long).unbind(0)
    mask = batch[i] == batch[j]
    i = i[mask]
    j = j[mask]
    return torch.stack([i, j])


class ScalarPlusWeightedCoulomb(OutputModel):
    """normal scalar output plus a coulomb interaction using predicted charges

    It assumes the input features are shape [N, hidden_channels + q_channels]
    where q_channels are the predicted partial charges (can be 1 or length N)
    If more than 1 the coulomb interaction is computed seperately for each charge
    channel and then a weighted mean is taken of the resulting energies.

    Implementation modifed from aimnetcentral
    """

    def __init__(
        self,
        hidden_channels,
        activation="silu",
        allow_prior_model=True,
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(ScalarPlusWeightedCoulomb, self).__init__(
            allow_prior_model=allow_prior_model,
            reduce_op=reduce_op,
            static_shapes=static_shapes,
        )
        self.hidden_channels = hidden_channels
        self.output_network = MLP(
            in_channels=hidden_channels,
            out_channels=1,
            hidden_channels=hidden_channels // 2,
            activation=activation,
            num_hidden_layers=kwargs.get("num_hidden_layers", 0),
            dtype=dtype,
        )

        self.q_dim = kwargs["q_dim"]
        self.num_interaction_layers = kwargs["num_layers"]
        self.layer_weights = kwargs["q_weights"]

        assert len(self.layer_weights) == self.num_interaction_layers + 1

        _w = torch.zeros((self.num_interaction_layers + 1, self.q_dim), dtype=dtype)
        for i in range(self.num_interaction_layers + 1):
            _w[i, :] = torch.tensor(self.layer_weights[i], dtype=dtype)

        _w = _w.flatten()

        self.register_buffer("qweights", _w)

        # from ase.units from aimnet2
        _Hartree = 27.211386024367243
        _half_Hartree = 0.5 * _Hartree
        _Bohr = 0.5291772105638411
        self._factor = _half_Hartree * _Bohr

        self.edge_index = torch.empty(2, 0)

        self.reset_parameters()

    def reset_parameters(self):
        self.output_network.reset_parameters()

    def pre_reduce(self, x, v: Optional[torch.Tensor], z, pos, batch):
        # we assume the charges have been added at the end of x!
        charges = x[:, self.hidden_channels :]

        if not self.static_shapes:  # check we have the expected number of charges
            assert charges.shape[1] == self.qweights.shape[0]

        x = x[:, : self.hidden_channels]
        x = self.output_network(x)  # energy per atom from main MLPs

        # torch.compile and torch.export don't support .item() calls during tracing
        # The model should be warmed up before compilation to set the correct dim_size
        if torch.compiler.is_compiling():
            assert self.edge_index.shape[1] > 0

        elif torch.jit.is_scripting():
            # TorchScript doesn't support torch.cuda.is_current_stream_capturing()
            # For CPU, always update dim_size (no CUDA graphs on CPU)
            # For CUDA with static_shapes, only update once (first call sets dim_size for CUDA graph capture)
            # For CUDA without static_shapes, always update (dynamic batch sizes)
            if not x.is_cuda or not self.static_shapes or self.edge_index.shape[1] == 0:
                self.edge_index = _triu_indices(x, batch)
        else:
            is_capturing = x.is_cuda and torch.cuda.is_current_stream_capturing()
            if not x.is_cuda or not is_capturing:
                self.edge_index = _triu_indices(x, batch)

            if is_capturing:
                assert (
                    self.edge_index is not None
                ), "Warming up is needed before capturing the model into a CUDA graph"
                warn(
                    "CUDA graph capture will lock the batch to the current number of samples ({}). Changing this will result in a crash".format(
                        self.dim_size
                    )
                )

        q_i = charges[self.edge_index[0]]
        q_j = charges[self.edge_index[1]]

        d_ij = torch.linalg.norm(
            pos[self.edge_index[0]] - pos[self.edge_index[1]], dim=-1
        )

        # charge product
        q_ij = q_i * q_j

        # short range damping, hyperparameter is from aimnet2
        fc = 1.0 - _exp_cutoff(d_ij, 4.6)

        # per edge Coulomb energy with short range damping
        e_ij = fc.unsqueeze(-1) * q_ij / d_ij.unsqueeze(-1)

        # scale with Coulomb constant in eV and Angstrom unit system!
        e_ij = self._factor * e_ij

        # now do the weighted mean over features
        e_ij = torch.sum(e_ij * self.qweights.unsqueeze(0), dim=-1) / torch.sum(
            self.qweights
        )

        # sum into nodes
        e_i = torch.zeros(x.shape[0], device=x.device, dtype=x.dtype)
        e_i = e_i.index_add(0, self.edge_index[0], e_ij)
        e_i = e_i.index_add(0, self.edge_index[1], e_ij)

        # add to normal per atom output with correct shape
        x = x + e_i.unsqueeze(-1)

        return x



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

from torch import nn, Tensor
from typing import Optional, Dict


class BasePrior(nn.Module):
    r"""Base class for prior models.
    Derive this class to make custom prior models, which take some arguments and a dataset as input.
    As an example, have a look at the `torchmdnet.priors.atomref.Atomref` prior.
    """

    def __init__(self, dataset=None):
        super().__init__()

    def get_init_args(self):
        r"""A function that returns all required arguments to construct a prior object.
        The values should be returned inside a dict with the keys being the arguments' names.
        All values should also be saveable in a .yaml file as this is used to reconstruct the
        prior model from a checkpoint file.
        """
        return {}

    def pre_reduce(self, x, z, pos, batch, extra_args: Optional[Dict[str, Tensor]]):
        r"""Pre-reduce method of the prior model.

        Args:
            x (torch.Tensor): scalar atom-wise predictions from the model.
            z (torch.Tensor): atom types of all atoms.
            pos (torch.Tensor): 3D atomic coordinates.
            batch (torch.Tensor): tensor containing the sample index for each atom.
            extra_args (dict): any addition fields provided by the dataset

        Returns:
            torch.Tensor: updated scalar atom-wise predictions
        """
        return x

    def post_reduce(
        self,
        y,
        z,
        pos,
        batch,
        box: Optional[Tensor],
        extra_args: Optional[Dict[str, Tensor]],
    ):
        r"""Post-reduce method of the prior model.

        Args:
            y (torch.Tensor): scalar molecule-wise predictions from the model.
            z (torch.Tensor): atom types of all atoms.
            pos (torch.Tensor): 3D atomic coordinates.
            batch (torch.Tensor): tensor containing the sample index for each atom.
            box (Optional[torch.Tensor]): box vectors of the system.
            extra_args (dict): any addition fields provided by the dataset

        Returns:
            torch.Tensor: updated scalar molecular-wise predictions
        """
        return y



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

# Place here any short extensions to torch that you want to use in your code.
# The extensions present in extensions.cpp will be automatically compiled in setup.py and loaded here.
# The extensions will be available under torch.ops.torchmdnet_extensions, but you can add wrappers here to make them more convenient to use.
# Place here too any meta registrations for your extensions if required.

import torch
from torch import Tensor
from typing import Tuple
import logging
from torchmdnet.extensions.neighbors import torch_neighbor_bruteforce

try:
    import triton
    from torchmdnet.extensions.triton_neighbors import triton_neighbor_pairs

    HAS_TRITON = True
except ImportError:
    HAS_TRITON = False


logger = logging.getLogger(__name__)

__all__ = ["get_neighbor_pairs_kernel"]


def get_neighbor_pairs_kernel(
    strategy: str,
    positions: Tensor,
    batch: Tensor,
    box_vectors: Tensor,
    use_periodic: bool,
    cutoff_lower: float,
    cutoff_upper: float,
    max_num_pairs: int,
    loop: bool,
    include_transpose: bool,
    num_cells: int,
) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
    """Computes the neighbor pairs for a given set of atomic positions.
    The list is generated as a list of pairs (i,j) without any enforced ordering.
    The list is padded with -1 to the maximum number of pairs.

    Parameters
    ----------
    strategy : str
        Strategy to use for computing the neighbor list. Can be one of :code:`["brute", "cell"]`.
    positions : Tensor
        A tensor with shape (N, 3) representing the atomic positions.
    batch : Tensor
        A tensor with shape (N,). Specifies the batch for each atom.
    box_vectors : Tensor
        The vectors defining the periodic box with shape `(3, 3)` or `(max(batch)+1, 3, 3)` if a different box is used for each sample.
    use_periodic : bool
        Whether to apply periodic boundary conditions.
    cutoff_lower : float
        Lower cutoff for the neighbor list.
    cutoff_upper : float
        Upper cutoff for the neighbor list.
    max_num_pairs : int
        Maximum number of pairs to store.
    loop : bool
        Whether to include self-interactions.
    include_transpose : bool
        Whether to include the transpose of the neighbor list (pair i,j and pair j,i).
    num_cells : int
        The number of cells in the grid if using the cell strategy.
    Returns
    -------
    neighbors : Tensor
        List of neighbors for each atom. Shape (2, max_num_pairs).
    distances : Tensor
        List of distances for each atom. Shape (max_num_pairs,).
    distance_vecs : Tensor
        List of distance vectors for each atom. Shape (max_num_pairs, 3).
    num_pairs : Tensor
        The number of pairs found.
    """
    if torch.jit.is_scripting() or not positions.is_cuda:

        return torch_neighbor_bruteforce(
            positions,
            batch=batch,
            box_vectors=box_vectors,
            use_periodic=use_periodic,
            cutoff_lower=cutoff_lower,
            cutoff_upper=cutoff_upper,
            max_num_pairs=max_num_pairs,
            loop=loop,
            include_transpose=include_transpose,
        )

    if not HAS_TRITON:
        logger.warning(
            "Triton is not available, using torch version of the neighbor pairs kernel."
        )
        return torch_neighbor_bruteforce(
            positions,
            batch=batch,
            box_vectors=box_vectors,
            use_periodic=use_periodic,
            cutoff_lower=cutoff_lower,
            cutoff_upper=cutoff_upper,
            max_num_pairs=max_num_pairs,
            loop=loop,
            include_transpose=include_transpose,
        )

    return triton_neighbor_pairs(
        strategy,
        positions=positions,
        batch=batch,
        box_vectors=box_vectors,
        use_periodic=use_periodic,
        cutoff_lower=cutoff_lower,
        cutoff_upper=cutoff_upper,
        max_num_pairs=max_num_pairs,
        loop=loop,
        include_transpose=include_transpose,
        num_cells=num_cells,
    )



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import yaml
import argparse
import numpy as np
import torch
from os.path import dirname, join, exists
from lightning_utilities.core.rank_zero import rank_zero_warn
import functools
import warnings

# fmt: off
# Atomic masses are based on:
#
#   Meija, J., Coplen, T., Berglund, M., et al. (2016). Atomic weights of
#   the elements 2013 (IUPAC Technical Report). Pure and Applied Chemistry,
#   88(3), pp. 265-291. Retrieved 30 Nov. 2016,
#   from doi:10.1515/pac-2015-0305
#
# Standard atomic weights are taken from Table 1: "Standard atomic weights
# 2013", with the uncertainties ignored.
# For hydrogen, helium, boron, carbon, nitrogen, oxygen, magnesium, silicon,
# sulfur, chlorine, bromine and thallium, where the weights are given as a
# range the "conventional" weights are taken from Table 3 and the ranges are
# given in the comments.
# The mass of the most stable isotope (in Table 4) is used for elements
# where there the element has no stable isotopes (to avoid NaNs): Tc, Pm,
# Po, At, Rn, Fr, Ra, Ac, everything after N
atomic_masses = np.array([
    1.0, 1.008, 4.002602, 6.94, 9.0121831,
    10.81, 12.011, 14.007, 15.999, 18.998403163,
    20.1797, 22.98976928, 24.305, 26.9815385, 28.085,
    30.973761998, 32.06, 35.45, 39.948, 39.0983,
    40.078, 44.955908, 47.867, 50.9415, 51.9961,
    54.938044, 55.845, 58.933194, 58.6934, 63.546,
    65.38, 69.723, 72.63, 74.921595, 78.971,
    79.904, 83.798, 85.4678, 87.62, 88.90584,
    91.224, 92.90637, 95.95, 97.90721, 101.07,
    102.9055, 106.42, 107.8682, 112.414, 114.818,
    118.71, 121.76, 127.6, 126.90447, 131.293,
    132.90545196, 137.327, 138.90547, 140.116, 140.90766,
    144.242, 144.91276, 150.36, 151.964, 157.25,
    158.92535, 162.5, 164.93033, 167.259, 168.93422,
    173.054, 174.9668, 178.49, 180.94788, 183.84,
    186.207, 190.23, 192.217, 195.084, 196.966569,
    200.592, 204.38, 207.2, 208.9804, 208.98243,
    209.98715, 222.01758, 223.01974, 226.02541, 227.02775,
    232.0377, 231.03588, 238.02891, 237.04817, 244.06421,
    243.06138, 247.07035, 247.07031, 251.07959, 252.083,
    257.09511, 258.09843, 259.101, 262.11, 267.122,
    268.126, 271.134, 270.133, 269.1338, 278.156,
    281.165, 281.166, 285.177, 286.182, 289.19,
    289.194, 293.204, 293.208, 294.214,
])
# fmt: on

ATOMIC_NUMBERS = {
    "H": 1,
    "He": 2,
    "Li": 3,
    "Be": 4,
    "B": 5,
    "C": 6,
    "N": 7,
    "O": 8,
    "F": 9,
    "Ne": 10,
    "Na": 11,
    "Mg": 12,
    "Al": 13,
    "Si": 14,
    "P": 15,
    "S": 16,
    "Cl": 17,
    "Ar": 18,
    "K": 19,
    "Ca": 20,
    "Sc": 21,
    "Ti": 22,
    "V": 23,
    "Cr": 24,
    "Mn": 25,
    "Fe": 26,
    "Co": 27,
    "Ni": 28,
    "Cu": 29,
    "Zn": 30,
    "Ga": 31,
    "Ge": 32,
    "As": 33,
    "Se": 34,
    "Br": 35,
    "Kr": 36,
    "Rb": 37,
    "Sr": 38,
    "Y": 39,
    "Zr": 40,
    "Nb": 41,
    "Mo": 42,
    "Tc": 43,
    "Ru": 44,
    "Rh": 45,
    "Pd": 46,
    "Ag": 47,
    "Cd": 48,
    "In": 49,
    "Sn": 50,
    "Sb": 51,
    "Te": 52,
    "I": 53,
    "Xe": 54,
    "Cs": 55,
    "Ba": 56,
    "La": 57,
    "Ce": 58,
    "Pr": 59,
    "Nd": 60,
    "Pm": 61,
    "Sm": 62,
    "Eu": 63,
    "Gd": 64,
    "Tb": 65,
    "Dy": 66,
    "Ho": 67,
    "Er": 68,
    "Tm": 69,
    "Yb": 70,
    "Lu": 71,
    "Hf": 72,
    "Ta": 73,
    "W": 74,
    "Re": 75,
    "Os": 76,
    "Ir": 77,
    "Pt": 78,
    "Au": 79,
    "Hg": 80,
    "Tl": 81,
    "Pb": 82,
    "Bi": 83,
    "Po": 84,
    "At": 85,
    "Rn": 86,
    "Fr": 87,
    "Ra": 88,
    "Ac": 89,
    "Th": 90,
    "Pa": 91,
    "U": 92,
    "Np": 93,
    "Pu": 94,
    "Am": 95,
    "Cm": 96,
    "Bk": 97,
    "Cf": 98,
    "Es": 99,
    "Fm": 100,
    "Md": 101,
    "No": 102,
    "Lr": 103,
    "Rf": 104,
    "Db": 105,
    "Sg": 106,
    "Bh": 107,
    "Hs": 108,
    "Mt": 109,
    "Ds": 110,
    "Rg": 111,
    "Cn": 112,
    "Nh": 113,
    "Fl": 114,
    "Mc": 115,
    "Lv": 116,
    "Ts": 117,
    "Og": 118,
}


def train_val_test_split(dset_len, train_size, val_size, test_size, seed, order=None):
    assert (train_size is None) + (val_size is None) + (
        test_size is None
    ) <= 1, "Only one of train_size, val_size, test_size is allowed to be None."
    is_float = (
        isinstance(train_size, float),
        isinstance(val_size, float),
        isinstance(test_size, float),
    )

    train_size = round(dset_len * train_size) if is_float[0] else train_size
    val_size = round(dset_len * val_size) if is_float[1] else val_size
    test_size = round(dset_len * test_size) if is_float[2] else test_size

    if train_size is None:
        train_size = dset_len - val_size - test_size
    elif val_size is None:
        val_size = dset_len - train_size - test_size
    elif test_size is None:
        test_size = dset_len - train_size - val_size

    if train_size + val_size + test_size > dset_len:
        if is_float[2]:
            test_size -= 1
        elif is_float[1]:
            val_size -= 1
        elif is_float[0]:
            train_size -= 1

    assert train_size >= 0 and val_size >= 0 and test_size >= 0, (
        f"One of training ({train_size}), validation ({val_size}) or "
        f"testing ({test_size}) splits ended up with a negative size."
    )

    total = train_size + val_size + test_size
    assert dset_len >= total, (
        f"The dataset ({dset_len}) is smaller than the "
        f"combined split sizes ({total})."
    )
    if total < dset_len:
        rank_zero_warn(f"{dset_len - total} samples were excluded from the dataset")

    idxs = np.arange(dset_len, dtype=int)
    if order is None:
        idxs = np.random.default_rng(seed).permutation(idxs)

    idx_train = idxs[:train_size]
    idx_val = idxs[train_size : train_size + val_size]
    idx_test = idxs[train_size + val_size : total]

    if order is not None:
        idx_train = [order[i] for i in idx_train]
        idx_val = [order[i] for i in idx_val]
        idx_test = [order[i] for i in idx_test]

    return np.array(idx_train), np.array(idx_val), np.array(idx_test)


def make_splits(
    dataset_len,
    train_size,
    val_size,
    test_size,
    seed,
    filename=None,
    splits=None,
    order=None,
):
    if splits is not None:
        splits = np.load(splits, allow_pickle=True)
        idx_train = splits["idx_train"]
        idx_val = splits["idx_val"]
        idx_test = splits["idx_test"]
    else:
        idx_train, idx_val, idx_test = train_val_test_split(
            dataset_len, train_size, val_size, test_size, seed, order
        )

    if filename is not None:
        np.savez(filename, idx_train=idx_train, idx_val=idx_val, idx_test=idx_test)

    return (
        torch.from_numpy(idx_train),
        torch.from_numpy(idx_val),
        torch.from_numpy(idx_test),
    )


class LoadFromFile(argparse.Action):
    # parser.add_argument('--file', type=open, action=LoadFromFile)
    def __call__(self, parser, namespace, values, option_string=None):
        if values.name.endswith("yaml") or values.name.endswith("yml"):
            with values as f:
                config = yaml.load(f, Loader=yaml.FullLoader)
            for key in config.keys():
                if key not in namespace:
                    raise ValueError(f"Unknown argument in config file: {key}")
            if (
                "load_model" in config
                and namespace.load_model is not None
                and config["load_model"] != namespace.load_model
            ):
                rank_zero_warn(
                    f"The load model argument was specified as a command line argument "
                    f"({namespace.load_model}) and in the config file ({config['load_model']}). "
                    f"Ignoring 'load_model' from the config file and loading {namespace.load_model}."
                )
                del config["load_model"]
            namespace.__dict__.update(config)
        else:
            raise ValueError("Configuration file must end with yaml or yml")


class LoadFromCheckpoint(argparse.Action):
    # parser.add_argument('--file', type=open, action=LoadFromFile)
    def __call__(self, parser, namespace, values, option_string=None):
        hparams_path = join(dirname(values), "hparams.yaml")
        if not exists(hparams_path):
            print(
                "Failed to locate the checkpoint's hparams.yaml file. Relying on command line args."
            )
            return
        with open(hparams_path, "r") as f:
            config = yaml.load(f, Loader=yaml.FullLoader)
        for key in config.keys():
            if key not in namespace and key != "prior_args":
                raise ValueError(f"Unknown argument in the model checkpoint: {key}")
        namespace.__dict__.update(config)
        namespace.__dict__.update(load_model=values)


def save_argparse(args, filename, exclude=None):
    import json

    if filename.endswith("yaml") or filename.endswith("yml"):
        if isinstance(exclude, str):
            exclude = [exclude]
        args = args.__dict__.copy()
        for exl in exclude:
            del args[exl]

        ds_arg = args.get("dataset_arg")
        if ds_arg is not None and isinstance(ds_arg, str):
            args["dataset_arg"] = json.loads(args["dataset_arg"])
        yaml.dump(args, open(filename, "w"))
    else:
        raise ValueError("Configuration file should end with yaml or yml")


def number(text):
    if text is None or text == "None":
        return None

    try:
        num_int = int(text)
    except ValueError:
        num_int = None
    num_float = float(text)

    if num_int == num_float:
        return num_int
    return num_float


class MissingEnergyException(Exception):
    pass


def write_as_hdf5(files, hdf5_dataset, tile_embed=True):
    """Transform the input numpy files to hdf5 format compatible with the HDF5 Dataset class.
    The input files to this function are the same as the ones required by the Custom dataset.
    Args:
        files (dict): Dictionary of numpy input files. Must contain "pos", "z" and at least one of "y" or "neg_dy".
        hdf5_dataset (string): Path to the output HDF5 dataset.
        tile_embed (bool): Whether to tile the embeddings to match the number of samples. Default: True
    Example:
        >>> files = {}
        >>> files["pos"] = sorted(glob.glob(join(tmpdir, "coords*")))
        >>> files["z"] = sorted(glob.glob(join(tmpdir, "embed*")))
        >>> files["y"] = sorted(glob.glob(join(tmpdir, "energy*")))
        >>> files["neg_dy"] = sorted(glob.glob(join(tmpdir, "forces*")))
        >>> write_as_hdf5(files, join(tmpdir, "test.hdf5"))
    """
    import h5py

    with h5py.File(hdf5_dataset, "w") as f:
        for i in range(len(files["pos"])):
            # Create a group for each file
            coord_data = np.load(files["pos"][i], mmap_mode="r")
            embed_data = np.load(files["z"][i], mmap_mode="r").astype(int)
            group = f.create_group(str(i))
            num_samples = coord_data.shape[0]
            group.create_dataset("pos", data=coord_data)
            if tile_embed:
                group.create_dataset(
                    "types", data=np.tile(embed_data, (num_samples, 1))
                )
            else:
                group.create_dataset("types", data=embed_data)
            if "y" in files:
                energy_data = np.load(files["y"][i], mmap_mode="r")
                group.create_dataset("energy", data=energy_data)
            if "neg_dy" in files:
                force_data = np.load(files["neg_dy"][i], mmap_mode="r")
                group.create_dataset("forces", data=force_data)


def deprecated_class(cls):
    """Decorator to mark classes as deprecated."""
    orig_init = cls.__init__

    @functools.wraps(orig_init)
    def wrapped_init(self, *args, **kwargs):
        warnings.simplefilter(
            "always", DeprecationWarning
        )  # ensure all deprecation warnings are shown
        warnings.warn(
            f"{cls.__name__} is deprecated and will be removed in a future version.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        orig_init(self, *args, **kwargs)

    cls.__init__ = wrapped_init
    return cls


def check_logs(log_dir):
    import os
    import time

    metr_file_path = os.path.join(log_dir, "metrics.csv")
    if os.path.exists(metr_file_path):
        # we make a backup of the metrics file (rename)
        bckp_date = f'{time.strftime("%Y%m%d")}-{time.strftime("%H%M%S")}'
        os.rename(metr_file_path, metr_file_path.replace(".csv", f"_{bckp_date}.csv"))
    return



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import hashlib
import h5py
import os
import torch as pt
from torchmdnet.datasets.memdataset import MemmappedDataset
from torch_geometric.data import Data, Dataset
from tqdm import tqdm


class Ace(MemmappedDataset):
    """The ACE dataset.

    This dataset is sourced from HDF5 files.

    Mandatory HDF5 file attributes:

    - `layout`: Must be set to `Ace`.
    - `layout_version`: Can be `1.0` or `2.0`.
    - `name`: Name of the dataset.

    For `layout_version` 1.0:

    - Files can contain multiple molecule groups directly under the root.
    - Each molecule group contains:

      - `atomic_numbers`: Atomic numbers of the atoms.
      - `formal_charges`: Formal charges of the atoms. The sum is the molecule's total charge. Units: electron charges.
      - `conformations` subgroup: This subgroup has individual conformation groups, each with datasets for different properties of the conformation.

    For `layout_version` 2.0:

    - Files contain a single root group (e.g., a 'master molecule group').
    - Within this root group, there can be multiple molecule groups.
    - Each molecule group contains:

      - `atomic_numbers`: Atomic numbers of the atoms.
      - `formal_charges`: Formal charges of the atoms.
      - Datasets for multiple conformations directly, without individual conformation groups.


    Each conformation group (version 1.0) or molecule group (version 2.0) should have the following datasets:

    - `positions`: Atomic positions. Units: Angstrom.
    - `forces`: Forces on the atoms. Units: eV/Å.
    - `partial_charges`: Atomic partial charges. Units: electron charges.
    - `dipole_moment` (version 1.0) or `dipole_moments` (version 2.0): Dipole moment (a vector of three components). Units: e*Å.
    - `formation_energy` (version 1.0) or `formation_energies` (version 2.0): Formation energy. Units: eV.

    Each dataset should also have an `units` attribute specifying its units (i.e., `Å`, `eV`, `e*Å`).

    Note that version 2.0 is more efficient than 1.0.

    Args:
        root (string, optional): Root directory where the dataset should be saved.
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version.
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version.
        pre_filter (callable, optional): A function that takes in an :obj:`torch_geometric.data.Data` object and returns a boolean value, indicating whether the data object should be included in the final dataset.
        paths (string or list): Path to the HDF5 files or directory containing the HDF5 files.
        max_gradient (float, optional): Maximum gradient norm. Samples with larger gradients are discarded.
        subsample_molecules (int, optional): Subsample molecules. Only every `subsample_molecules` molecule is used.

    Examples::
        >>> import numpy as np
        >>> from torchmdnet.datasets import Ace
        >>> import h5py
        >>> # Version 1.0 example
        >>> with h5py.File("molecule.h5", 'w') as f:
        ...     f.attrs["layout"] = "Ace"
        ...     f.attrs["layout_version"] = "1.0"
        ...     f.attrs["name"] = "sample_molecule_data"
        ...     for m in range(3):  # Three molecules
        ...         mol = f.create_group(f"mol_{m+1}")
        ...         mol["atomic_numbers"] = [1, 6, 8]  # H, C, O
        ...         mol["formal_charges"] = [0, 0, 0]  # Neutral charges
        ...         confs = mol.create_group("conformations")
        ...         for i in range(2):  # Two conformations
        ...             conf = confs.create_group(f"conf_{i+1}")
        ...             conf["positions"] = np.random.random((3, 3))
        ...             conf["positions"].attrs["units"] = "Å"
        ...             conf["formation_energy"] = np.random.random()
        ...             conf["formation_energy"].attrs["units"] = "eV"
        ...             conf["forces"] = np.random.random((3, 3))
        ...             conf["forces"].attrs["units"] = "eV/Å"
        ...             conf["partial_charges"] = np.random.random(3)
        ...             conf["partial_charges"].attrs["units"] = "e"
        ...             conf["dipole_moment"] = np.random.random(3)
        ...             conf["dipole_moment"].attrs["units"] = "e*Å"
        >>> dataset = Ace(root=".", paths="molecule.h5")
        >>> len(dataset)
        6
        >>> dataset = Ace(root=".", paths=["molecule.h5", "molecule.h5"])
        >>> len(dataset)
        12
        >>> # Version 2.0 example
        >>> with h5py.File("molecule_v2.h5", 'w') as f:
        ...     f.attrs["layout"] = "Ace"
        ...     f.attrs["layout_version"] = "2.0"
        ...     f.attrs["name"] = "sample_molecule_data_v2"
        ...     master_mol_group = f.create_group("master_molecule_group")
        ...     for m in range(3):  # Three molecules
        ...         mol = master_mol_group.create_group(f"mol_{m+1}")
        ...         mol["atomic_numbers"] = [1, 6, 8]  # H, C, O
        ...         mol["formal_charges"] = [0, 0, 0]  # Neutral charges
        ...         mol["positions"] = np.random.random((2, 3, 3))  # Two conformations
        ...         mol["positions"].attrs["units"] = "Å"
        ...         mol["formation_energies"] = np.random.random(2)
        ...         mol["formation_energies"].attrs["units"] = "eV"
        ...         mol["forces"] = np.random.random((2, 3, 3))
        ...         mol["forces"].attrs["units"] = "eV/Å"
        ...         mol["partial_charges"] = np.random.random((2, 3))
        ...         mol["partial_charges"].attrs["units"] = "e"
        ...         mol["dipole_moment"] = np.random.random((2, 3))
        ...         mol["dipole_moment"].attrs["units"] = "e*Å"
        >>> dataset_v2 = Ace(root=".", paths="molecule_v2.h5")
        >>> len(dataset_v2)
        6
        >>> dataset_v2 = Ace(root=".", paths=["molecule_v2.h5", "molecule_v2.h5"])
        >>> len(dataset_v2)
        12
    """

    def __init__(
        self,
        root=None,
        transform=None,
        pre_transform=None,
        pre_filter=None,
        paths=None,
        max_gradient=None,
        subsample_molecules=1,
    ):
        assert isinstance(paths, (str, list))

        arg_hash = f"{paths}{max_gradient}{subsample_molecules}"
        arg_hash = hashlib.md5(arg_hash.encode()).hexdigest()
        self.name = f"{self.__class__.__name__}-{arg_hash}"
        self.paths = paths
        self.max_gradient = max_gradient
        self.subsample_molecules = int(subsample_molecules)

        props = ["y", "neg_dy", "q", "pq", "dp"]
        super().__init__(
            root,
            transform,
            pre_transform,
            pre_filter,
            properties=tuple(props),
        )

    @property
    def raw_paths(self):
        paths_init = self.paths if isinstance(self.paths, list) else [self.paths]
        paths = []
        for path in paths_init:
            if os.path.isfile(path):
                paths.append(path)
                continue

            if os.path.isdir(path):
                for file_ in os.listdir(path):
                    if file_.endswith(".h5"):
                        paths.append(os.path.join(path, file_))
                continue

            raise RuntimeError(f"{path} is neither a directory nor a file")

        return paths

    @staticmethod
    def _load_confs_1_0(mol, n_atoms):
        for conf in mol["conformations"].values():
            # Skip failed calculations
            if "formation_energy" not in conf:
                continue

            assert conf["positions"].attrs["units"] == "Å"
            pos = pt.tensor(conf["positions"][...], dtype=pt.float32)
            assert pos.shape == (n_atoms, 3)

            assert conf["formation_energy"].attrs["units"] == "eV"
            y = pt.tensor(conf["formation_energy"][()], dtype=pt.float64)
            assert y.shape == ()

            assert conf["forces"].attrs["units"] == "eV/Å"
            neg_dy = pt.tensor(conf["forces"][...], dtype=pt.float32)
            assert neg_dy.shape == pos.shape

            assert conf["partial_charges"].attrs["units"] == "e"
            pq = pt.tensor(conf["partial_charges"][:], dtype=pt.float32)
            assert pq.shape == (n_atoms,)

            assert conf["dipole_moment"].attrs["units"] == "e*Å"
            dp = pt.tensor(conf["dipole_moment"][:], dtype=pt.float32)
            assert dp.shape == (3,)

            yield pos, y, neg_dy, pq, dp

    @staticmethod
    def _load_confs_2_0(mol, n_atoms):
        assert mol["positions"].attrs["units"] == "Å"
        all_pos = pt.tensor(mol["positions"][...], dtype=pt.float32)
        n_confs = all_pos.shape[0]
        assert all_pos.shape == (n_confs, n_atoms, 3)

        assert mol["formation_energies"].attrs["units"] == "eV"
        all_y = pt.tensor(mol["formation_energies"][:], dtype=pt.float64)
        assert all_y.shape == (n_confs,)

        assert mol["forces"].attrs["units"] == "eV/Å"
        all_neg_dy = pt.tensor(mol["forces"][...], dtype=pt.float32)
        assert all_neg_dy.shape == all_pos.shape

        assert mol["partial_charges"].attrs["units"] == "e"
        all_pq = pt.tensor(mol["partial_charges"][...], dtype=pt.float32)
        assert all_pq.shape == (n_confs, n_atoms)

        assert mol["dipole_moments"].attrs["units"] == "e*Å"
        all_dp = pt.tensor(mol["dipole_moments"][...], dtype=pt.float32)
        assert all_dp.shape == (n_confs, 3)

        for pos, y, neg_dy, pq, dp in zip(all_pos, all_y, all_neg_dy, all_pq, all_dp):
            # Skip failed calculations
            if y.isnan():
                continue

            yield pos, y, neg_dy, pq, dp

    def sample_iter(self, mol_ids=False):
        assert self.subsample_molecules > 0

        for i_path, path in tqdm(enumerate(self.raw_paths), desc="Files"):
            h5 = h5py.File(path)
            assert h5.attrs["layout"] == "Ace"
            version = h5.attrs["layout_version"]

            mols = None
            load_confs = None
            if version == "1.0":
                assert "name" in h5.attrs
                mols = h5.items()
                load_confs = self._load_confs_1_0
            elif version == "2.0":
                assert len(h5.keys()) == 1
                mols = list(h5.values())[0].items()
                load_confs = self._load_confs_2_0
            else:
                raise RuntimeError(f"Unsupported layout version: {version}")

            # Iterate over the molecules
            for i_mol, (mol_id, mol) in tqdm(
                enumerate(mols),
                desc="Molecules",
                total=len(mols),
                leave=False,
            ):
                # Subsample molecules
                if i_mol % self.subsample_molecules != 0:
                    continue

                z = pt.tensor(mol["atomic_numbers"], dtype=pt.long)
                fq = pt.tensor(mol["formal_charges"], dtype=pt.long)
                q = fq.sum()

                for i_conf, (pos, y, neg_dy, pq, dp) in enumerate(
                    load_confs(mol, n_atoms=len(z))
                ):
                    # Skip samples with large forces
                    if self.max_gradient:
                        if neg_dy.norm(dim=1).max() > float(self.max_gradient):
                            continue

                    # Create a sample
                    args = dict(
                        z=z, pos=pos, y=y.view(1, 1), neg_dy=neg_dy, q=q, pq=pq, dp=dp
                    )
                    if mol_ids:
                        args["i_path"] = i_path
                        args["mol_id"] = mol_id
                        args["i_conf"] = i_conf

                    data = Data(**args)

                    if self.pre_filter is not None and not self.pre_filter(data):
                        continue

                    if self.pre_transform is not None:
                        data = self.pre_transform(data)

                    yield data


def download_gitea_dataset(path, tmpdir):
    try:
        from git import Repo
    except ImportError:
        raise ImportError(
            "Could not import GitPython library. Please install it first with `pip install GitPython`"
        )

    assert path.startswith("ssh://")

    # Parse the gitea URL
    pieces = path.split("/")
    repo_url = "/".join(pieces[:5])
    user = pieces[3]
    repo_name = pieces[4]
    file_name = pieces[-1]
    branch = "main"
    commit = None
    if "branch" in pieces:
        branch = pieces[pieces.index("branch") + 1]
    if "commit" in pieces:
        commit = pieces[pieces.index("commit") + 1]

    outdir = os.path.join(tmpdir, f"{user}_{repo_name}")
    if not os.path.exists(outdir):
        repo = Repo.clone_from(repo_url, outdir, no_checkout=True)
    else:
        repo = Repo(outdir)

    origin = repo.remotes.origin
    origin.pull()
    if commit is not None:
        repo.git.checkout(commit)
    else:
        repo.git.checkout(branch)

    return os.path.join(outdir, file_name)


class AceHF(Dataset):
    def __init__(
        self, root="parquet", paths=None, split="train", max_gradient=None
    ) -> None:
        from datasets import load_dataset
        import numpy as np

        # Handle gitea parquet datasets
        newpaths = paths.copy()
        for i, path in enumerate(paths):
            if "gitea" in path:
                newpaths[i] = download_gitea_dataset(path, "/tmp")

        self.dataset = load_dataset(root, data_files=newpaths, split=split)
        if max_gradient is not None:

            def _filter(x):
                if np.isnan(x["forces"]).any() or np.isnan(x["formation_energy"]).any():
                    return False
                return np.max(np.linalg.norm(x["forces"], axis=1)) < max_gradient

            self.dataset = self.dataset.filter(
                _filter, desc="Filtering", num_proc=os.cpu_count() // 2
            )
        self.dataset = self.dataset.with_format("torch")

    def __len__(self):
        return self.dataset.num_rows

    def __getitem__(self, idx):
        """Gets the data object at index :obj:`idx`.

        The data object contains the following attributes:

            - :obj:`z`: Atomic numbers of the atoms.
            - :obj:`pos`: Positions of the atoms.
            - :obj:`y`: Formation energy of the molecule.
            - :obj:`neg_dy`: Forces on the atoms.
            - :obj:`q`: Total charge of the molecule.
            - :obj:`pq`: Partial charges of the atoms.
            - :obj:`dp`: Dipole moment of the molecule.

        Args:
            idx (int): Index of the data object.

        Returns:
            :obj:`torch_geometric.data.Data`: The data object.
        """
        data = self.dataset[int(idx)]
        return Data(
            z=data["atomic_numbers"],
            pos=data["positions"],
            y=data["formation_energy"].view(1, 1),
            neg_dy=data["forces"],
            q=sum(data["formal_charges"]),
            pq=data["partial_charges"],
            dp=data["dipole_moment"],
        )



import torch
from typing import Optional, Tuple
from torch import Tensor, nn
from torchmdnet.models.utils import (
    CosineCutoff,
    OptimizedDistance,
    rbf_class_mapping,
    act_class_mapping,
    scatter,
    MLP,
)
from warnings import warn


from torchmdnet.models.tensornet import (
    vector_to_skewtensor,
    skewtensor_to_vector,
    decompose_tensor,
    I_to_tensor,
    tensor_norm,
    TensorEmbedding,
    scalar_message_passing,
    vector_message_passing,
    tensor_message_passing,
)


__all__ = ["TensorNet2"]


class ChargePredict(nn.Module):
    def __init__(self, hidden_channels, activation, q_dim=16, static_shapes=False):
        super(ChargePredict, self).__init__()
        self.q_dim = q_dim
        self.q_norm = nn.LayerNorm(3 * hidden_channels)
        self.q_mlp = MLP(3 * hidden_channels, 2 * q_dim, hidden_channels, activation, 1)
        self.static_shapes = static_shapes
        self.dim_size = 0

    def reset_parameters(self):
        self.q_norm.reset_parameters()
        self.q_mlp.reset_parameters()

    def mol_sum(self, x: Tensor, batch: Tensor) -> Tensor:
        # torch.compile and torch.export don't support .item() calls during tracing
        # The model should be warmed up before compilation to set the correct dim_size
        if torch.compiler.is_compiling():
            pass
        elif torch.jit.is_scripting():
            # TorchScript doesn't support torch.cuda.is_current_stream_capturing()
            # For CPU, always update dim_size (no CUDA graphs on CPU)
            # For CUDA with static_shapes, only update once (first call sets dim_size for CUDA graph capture)
            # For CUDA without static_shapes, always update (dynamic batch sizes)
            if not x.is_cuda or not self.static_shapes or self.dim_size == 0:
                self.dim_size = int(batch.max().item() + 1)
        else:
            is_capturing = x.is_cuda and torch.cuda.is_current_stream_capturing()
            if not x.is_cuda or not is_capturing:
                self.dim_size = int(batch.max().item() + 1)
            if is_capturing:
                assert (
                    self.dim_size > 0
                ), "Warming up is needed before capturing the model into a CUDA graph"
                warn(
                    "CUDA graph capture will lock the batch to the current number of samples ({}). Changing this will result in a crash".format(
                        self.dim_size
                    )
                )

        if self.dim_size == 1:
            # do the direct sum
            out = x.sum(dim=0).unsqueeze(0)
        else:
            out = torch.zeros(
                self.dim_size, x.shape[-1], device=x.device, dtype=x.dtype
            )
            out = out.index_add(0, batch, x)

        return out

    def qeq(self, old_charges, f, batch, Q) -> Tensor:

        # if we are using static shapes we can skip the last dummy atom
        if self.static_shapes:
            f = f[:-1]
            old_charges = old_charges[:-1]
            Q = Q[:-1]  # Q is already per atom

        f_u = f**2

        epsilon = 1.0e-6

        F_u = self.mol_sum(f_u, batch)
        F_u = F_u + epsilon
        Q_u = self.mol_sum(old_charges, batch)

        # Q is already per atom
        dQ = Q.unsqueeze(-1) - Q_u[batch]

        F_u = F_u[batch]

        _f = f_u / F_u
        new_charges = old_charges + _f * dQ

        if self.static_shapes:  # add back the dummy atom so shapes match
            new_charges = torch.cat(
                (
                    new_charges,
                    torch.zeros(
                        (1, new_charges.shape[1]),
                        device=new_charges.device,
                        dtype=new_charges.dtype,
                    ),
                )
            )

        return new_charges

    def forward(self, X, batch, Q):
        I, A, S = decompose_tensor(X)
        _charges_f = self.q_mlp(
            self.q_norm(torch.cat((I, tensor_norm(A), tensor_norm(S)), dim=-1))
        )
        charges, _f = _charges_f[:, : self.q_dim], _charges_f[:, self.q_dim :]
        charges = self.qeq(charges, _f, batch, Q)

        return charges


class TensorNet2(nn.Module):
    r"""TensorNet with Neutral Charge Equilibration
    
    
    Combines AIMNet2 style neutral charge equilibration procedure with TensorNet
    
    
    TensorNet: Cartesian Tensor Representations for Efficient Learning of Molecular Potentials; G. Simeon and G. de Fabritiis.
    NeurIPS 2023.

    This function optionally supports periodic boundary conditions with arbitrary triclinic boxes.
    For a given cutoff, :math:`r_c`, the box vectors :math:`\vec{a},\vec{b},\vec{c}` must satisfy certain requirements:

    .. math::

      \begin{align*}
      a_y = a_z = b_z &= 0 \\
      a_x, b_y, c_z &\geq 2 r_c \\
      a_x &\geq 2  b_x \\
      a_x &\geq 2  c_x \\
      b_y &\geq 2  c_y
      \end{align*}

    These requirements correspond to a particular rotation of the system and reduced form of the vectors, as well as the requirement that the cutoff be no larger than half the box width.

    Args:
        hidden_channels (int, optional): Hidden embedding size.
            (default: :obj:`128`)
        q_dim (int, optional): Charge channel size.
            (default: :obj:`16`)
        num_layers (int, optional): The number of interaction layers.
            (default: :obj:`2`)
        num_rbf (int, optional): The number of radial basis functions :math:`\mu`.
            (default: :obj:`32`)
        rbf_type (string, optional): The type of radial basis function to use.
            (default: :obj:`"expnorm"`)
        trainable_rbf (bool, optional): Whether to train RBF parameters with
            backpropagation. (default: :obj:`False`)
        activation (string, optional): The type of activation function to use.
            (default: :obj:`"silu"`)
        cutoff_lower (float, optional): Lower cutoff distance for interatomic interactions.
            (default: :obj:`0.0`)
        cutoff_upper (float, optional): Upper cutoff distance for interatomic interactions.
            (default: :obj:`4.5`)
        max_z (int, optional): Maximum atomic number. Used for initializing embeddings.
            (default: :obj:`128`)
        max_num_neighbors (int, optional): Maximum number of neighbors to return for a
            given node/atom when constructing the molecular graph during forward passes.
            (default: :obj:`64`)
        equivariance_invariance_group (string, optional): Group under whose action on input
            positions internal tensor features will be equivariant and scalar predictions
            will be invariant. O(3) or SO(3).
            (default :obj:`"O(3)"`)
        box_vecs (Tensor, optional):
            The vectors defining the periodic box.  This must have shape `(3, 3)`,
            where `box_vectors[0] = a`, `box_vectors[1] = b`, and `box_vectors[2] = c`.
            If this is omitted, periodic boundary conditions are not applied.
            (default: :obj:`None`)
        static_shapes (bool, optional): Whether to enforce static shapes.
            Makes the model CUDA-graph compatible if check_errors is set to False.
            (default: :obj:`True`)
        check_errors (bool, optional): Whether to check for errors in the distance module.
            (default: :obj:`True`)
        output_charges (bool, optional): Whether charges should be output. The current 
            output is done by appending to the node features x that will be passed to the output 
            model. If True the output model must be compatible with the output charges.
            (default: :obj:`False`)
    """

    def __init__(
        self,
        hidden_channels=128,
        q_dim=16,
        num_layers=2,
        num_rbf=32,
        rbf_type="expnorm",
        trainable_rbf=False,
        activation="silu",
        cutoff_lower=0,
        cutoff_upper=4.5,
        max_num_neighbors=64,
        max_z=128,
        equivariance_invariance_group="O(3)",
        static_shapes=True,
        check_errors=True,
        dtype=torch.float32,
        box_vecs=None,
        output_charges=False,
    ):
        super(TensorNet2, self).__init__()

        self.q_dim = q_dim
        self.output_charges = output_charges

        assert rbf_type in rbf_class_mapping, (
            f'Unknown RBF type "{rbf_type}". '
            f'Choose from {", ".join(rbf_class_mapping.keys())}.'
        )
        assert activation in act_class_mapping, (
            f'Unknown activation function "{activation}". '
            f'Choose from {", ".join(act_class_mapping.keys())}.'
        )

        assert equivariance_invariance_group in ["O(3)", "SO(3)"], (
            f'Unknown group "{equivariance_invariance_group}". '
            f"Choose O(3) or SO(3)."
        )
        self.hidden_channels = hidden_channels
        self.equivariance_invariance_group = equivariance_invariance_group
        self.num_layers = num_layers
        self.num_rbf = num_rbf
        self.rbf_type = rbf_type
        self.activation = activation
        self.cutoff_lower = cutoff_lower
        self.cutoff_upper = cutoff_upper
        act_class = act_class_mapping[activation]
        self.distance_expansion = rbf_class_mapping[rbf_type](
            cutoff_lower, cutoff_upper, num_rbf, trainable_rbf
        )
        self.tensor_embedding = TensorEmbedding(
            hidden_channels,
            num_rbf,
            act_class,
            cutoff_lower,
            cutoff_upper,
            trainable_rbf,
            max_z,
            dtype,
        )

        self.charge_predict_0 = ChargePredict(
            hidden_channels, activation, self.q_dim, static_shapes
        )

        self.layers = nn.ModuleList()
        self.charge_predicts = nn.ModuleList()
        if num_layers != 0:
            for l in range(num_layers):
                _layer = Interaction(
                    num_rbf,
                    hidden_channels,
                    self.q_dim,
                    activation,
                    cutoff_lower,
                    cutoff_upper,
                    equivariance_invariance_group,
                    dtype,
                )
                self.layers.append(_layer)
                _charge_predict = ChargePredict(
                    hidden_channels, activation, self.q_dim, static_shapes
                )
                self.charge_predicts.append(_charge_predict)

        self.linear = nn.Linear(3 * hidden_channels, hidden_channels, dtype=dtype)
        self.out_norm = nn.LayerNorm(3 * hidden_channels, dtype=dtype)
        self.act = act_class()
        # Resize to fit set to false ensures Distance returns a statically-shaped tensor of size max_num_pairs=pos.size*max_num_neigbors
        # negative max_num_pairs argument means "per particle"
        # long_edge_index set to False saves memory and spares some kernel launches by keeping neighbor indices as int32.
        self.static_shapes = static_shapes
        self.distance = OptimizedDistance(
            cutoff_lower,
            cutoff_upper,
            max_num_pairs=-max_num_neighbors,
            return_vecs=True,
            loop=True,
            check_errors=check_errors,
            resize_to_fit=not self.static_shapes,
            box=box_vecs,
            long_edge_index=True,
        )
        self.reset_parameters()

    def reset_parameters(self):
        self.tensor_embedding.reset_parameters()
        for layer in self.layers:
            layer.reset_parameters()
        self.linear.reset_parameters()
        self.out_norm.reset_parameters()
        self.charge_predict_0.reset_parameters()
        for layer in self.charge_predicts:
            layer.reset_parameters()

    def forward(
        self,
        z: Tensor,
        pos: Tensor,
        batch: Tensor,
        box: Optional[Tensor] = None,
        q: Optional[Tensor] = None,
        s: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Optional[Tensor], Tensor, Tensor, Tensor]:

        Q = q  # total molecule charge

        # Obtain graph, with distances and relative position vectors
        edge_index, edge_weight, edge_vec = self.distance(pos, batch, box)

        # This assert convinces TorchScript that edge_vec is a Tensor and not an Optional[Tensor]
        assert (
            edge_vec is not None
        ), "Distance module did not return directional information"
        # Total charge Q is a molecule-wise property. We transform it into an atom-wise property, with all atoms belonging to the same molecule being assigned the same charge q
        if Q is None:
            Q = torch.zeros_like(z, device=z.device, dtype=z.dtype)
        else:
            Q = Q[batch]
        zp = z
        if self.static_shapes:
            mask = (edge_index[0] < 0).unsqueeze(0).expand_as(edge_index)
            zp = torch.cat((z, torch.zeros(1, device=z.device, dtype=z.dtype)), dim=0)
            Q = torch.cat((Q, torch.zeros(1, device=Q.device, dtype=Q.dtype)), dim=0)
            # I trick the model into thinking that the masked edges pertain to the extra atom
            # WARNING: This can hurt performance if max_num_pairs >> actual_num_pairs
            edge_index = edge_index.masked_fill(mask, z.shape[0])
            edge_weight = edge_weight.masked_fill(mask[0], 0)
            edge_vec = edge_vec.masked_fill(
                mask[0].unsqueeze(-1).expand_as(edge_vec), 0
            )
        edge_attr = self.distance_expansion(edge_weight)
        mask = edge_index[0] == edge_index[1]
        # Normalizing edge vectors by their length can result in NaNs, breaking Autograd.
        # I avoid dividing by zero by setting the weight of self edges and self loops to 1
        edge_vec = edge_vec / edge_weight.masked_fill(mask, 1).unsqueeze(1)
        X = self.tensor_embedding(zp, edge_index, edge_weight, edge_vec, edge_attr)

        # from the initial embedding we compute some partial charges
        charges_0 = self.charge_predict_0(X, batch, Q)

        # we save them for the ouput layer
        charge_list = [charges_0]

        old_charges = charges_0  # intermediate so we can loop

        for l, (layer, charge_predictor) in enumerate(
            zip(self.layers, self.charge_predicts)
        ):
            # updated node tensor feratures
            X = layer(X, old_charges, edge_index, edge_weight, edge_attr)

            # predict new charges
            new_charges = charge_predictor(X, batch, Q)
            charge_list.append(new_charges)
            old_charges = new_charges

        # we will have (l+1 * qdim) charges for each node
        charges = torch.cat(charge_list, dim=-1)

        I, A, S = decompose_tensor(X)
        x = torch.cat((3 * I**2, tensor_norm(A), tensor_norm(S)), dim=-1)
        x = self.out_norm(x)
        x = self.act(self.linear((x)))

        # append charges for coulomb in output module
        if self.output_charges:
            x = torch.cat([x, charges], dim=-1)

        # Remove the extra atom
        if self.static_shapes:
            x = x[:-1]
        return x, None, z, pos, batch


class Interaction(nn.Module):
    """Interaction layer.

    :meta private:
    """

    def __init__(
        self,
        num_rbf,
        hidden_channels,
        q_dim,
        activation,
        cutoff_lower,
        cutoff_upper,
        equivariance_invariance_group,
        dtype=torch.float32,
    ):
        super(Interaction, self).__init__()

        self.hidden_channels = hidden_channels
        self.cutoff = CosineCutoff(cutoff_lower, cutoff_upper)
        self.linears_scalar = nn.ModuleList()
        self.linears_scalar.append(
            nn.Linear(num_rbf + 2 * q_dim, self.hidden_channels, bias=True, dtype=dtype)
        )
        self.linears_scalar.append(
            nn.Linear(
                self.hidden_channels, 2 * self.hidden_channels, bias=True, dtype=dtype
            )
        )
        self.linears_scalar.append(
            nn.Linear(
                2 * self.hidden_channels,
                3 * self.hidden_channels,
                bias=True,
                dtype=dtype,
            )
        )
        self.linears_tensor = nn.ModuleList()
        for _ in range(6):
            self.linears_tensor.append(
                nn.Linear(hidden_channels, self.hidden_channels, bias=False)
            )
        self.act = act_class_mapping[activation]()
        self.equivariance_invariance_group = equivariance_invariance_group
        self.reset_parameters()

    def reset_parameters(self):
        for linear in self.linears_scalar:
            linear.reset_parameters()
        for linear in self.linears_tensor:
            linear.reset_parameters()

    def forward(
        self,
        X: Tensor,
        charges: Tensor,
        edge_index: Tensor,
        edge_weight: Tensor,
        edge_attr: Tensor,
    ) -> Tensor:

        C = self.cutoff(edge_weight)

        # add the partial charges as edge features and then do a normal tensornet update
        charge_edges_i = charges.index_select(0, edge_index[0])
        charge_edges_j = charges.index_select(0, edge_index[1])
        x_edge_attr = torch.cat([edge_attr, charge_edges_i, charge_edges_j], dim=-1)

        for linear_scalar in self.linears_scalar:
            x_edge_attr = self.act(linear_scalar(x_edge_attr))
        x_edge_attr = (x_edge_attr * C.view(-1, 1)).reshape(
            x_edge_attr.shape[0], 3, self.hidden_channels
        )

        X = X / (tensor_norm(X) + 1)[..., None, None]
        I, A, S = decompose_tensor(X)
        I = self.linears_tensor[0](I)  # scalar
        A = self.linears_tensor[1](A.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)  # tensor
        S = self.linears_tensor[2](S.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)  # tensor
        Y = A + S + I_to_tensor(I)

        Im = scalar_message_passing(
            edge_index, x_edge_attr[..., 0, :], I, X.shape[0]
        )  # scalar
        Am_vec = vector_message_passing(
            edge_index,
            x_edge_attr[..., 1, :, None],
            skewtensor_to_vector(A),  # tensor -> vector
            X.shape[0],
        )  # vector
        Sm = tensor_message_passing(
            edge_index, x_edge_attr[..., 2, :, None, None], S, X.shape[0]
        )  # tensor
        msg = vector_to_skewtensor(Am_vec) + Sm + I_to_tensor(Im)

        if self.equivariance_invariance_group == "O(3)":
            A = torch.matmul(msg, Y)
            B = torch.matmul(Y, msg)
            I, A, S = decompose_tensor((A + B))
        if self.equivariance_invariance_group == "SO(3)":
            B = torch.matmul(Y, msg)
            I, A, S = decompose_tensor(2 * B)

        _X = A + S + I_to_tensor(I)
        normp1 = tensor_norm(_X) + 1
        I, A, S = I / normp1, A / normp1[..., None, None], S / normp1[..., None, None]
        I = self.linears_tensor[3](I)
        A = self.linears_tensor[4](A.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        S = self.linears_tensor[5](S.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        dX = A + S + I_to_tensor(I)
        X = X + dX + torch.matrix_power(dX, 2)
        return X
